from abc import abstractmethod
from typing import Any

import torch


class TorchTransformerMixin:
    """Mixin class for all transformers in scikit-learn.

    This mixin defines the following functionality:

    - a `fit_transform` method that delegates to `fit` and `transform`;
    - a `set_output` method to output `X` as a specific container type.

    If :term:`get_feature_names_out` is defined, then :class:`BaseEstimator` will
    automatically wrap `transform` and `fit_transform` to follow the `set_output`
    API. See the :ref:`developer_api_set_output` for details.

    :class:`OneToOneFeatureMixin` and
    :class:`ClassNamePrefixFeaturesOutMixin` are helpful mixins for
    defining :term:`get_feature_names_out`.

    Examples
    --------
    >>> import numpy as np
    >>> from sklearn.base import BaseEstimator, TransformerMixin
    >>> class MyTransformer(TransformerMixin, BaseEstimator):
    ...     def __init__(self, *, param=1):
    ...         self.param = param
    ...     def fit(self, X, y=None):
    ...         return self
    ...     def transform(self, X):
    ...         return np.full(shape=len(X), fill_value=self.param)
    >>> transformer = MyTransformer()
    >>> X = [[1, 2], [2, 3], [3, 4]]
    >>> transformer.fit_transform(X)
    array([1, 1, 1])
    """

    @abstractmethod
    def fit(self, X: torch.Tensor, **fit_kwargs: Any) -> "TorchTransformerMixin":
        """"""

    @abstractmethod
    def transform(self, X: torch.Tensor, **transform_kwargs: Any) -> torch.Tensor:
        """"""

    @abstractmethod
    def fit_transform(self, X: torch.Tensor, **kwargs: Any) -> torch.Tensor:
        """"""



import collections
from typing import List, Literal, OrderedDict

import torch

from .common import lazy_normalize


DistanceOptions = Literal["cosine", "euclidean"]
AffinityOptions = Literal["cosine", "rbf", "laplacian"]

# noinspection PyTypeChecker
DISTANCE_TO_AFFINITY: OrderedDict[DistanceOptions, List[AffinityOptions]] = collections.OrderedDict([
    ("cosine", ["cosine"]),
    ("euclidean", ["rbf", "laplacian"]),
])
# noinspection PyTypeChecker
AFFINITY_TO_DISTANCE: OrderedDict[AffinityOptions, DistanceOptions] = collections.OrderedDict(sum([
    [(affinity_type, distance_type) for affinity_type in affinity_types]
    for distance_type, affinity_types in DISTANCE_TO_AFFINITY.items()
], start=[]))



def to_euclidean(x: torch.Tensor, distance_type: DistanceOptions) -> torch.Tensor:
    if distance_type == "cosine":
        return lazy_normalize(x, p=2, dim=-1)
    elif distance_type == "euclidean":
        return x
    else:
        raise ValueError(f"to_euclidean not implemented for distance_type {distance_type}.")


def distance_from_features(
    features: torch.Tensor,
    features_B: torch.Tensor,
    distance_type: DistanceOptions,
):
    """Compute distance matrix from input features.
    Args:
        features (torch.Tensor): input features, shape (n_samples, n_features)
        features_B (torch.Tensor, optional): optional, if not None, compute affinity between two features
        distance_type (str): distance metric, 'cosine' (default) or 'euclidean', 'rbf'.
    Returns:
        (torch.Tensor): affinity matrix, shape (n_samples, n_samples)
    """
    # compute distance matrix from input features
    shape: torch.Size = features.shape[:-2]
    features = features.view((-1, *features.shape[-2:]))
    features_B = features_B.view((-1, *features_B.shape[-2:]))

    match distance_type:
        case "cosine":
            features = lazy_normalize(features, dim=-1)
            features_B = lazy_normalize(features_B, dim=-1)
            D = 1 - features @ features_B.mT
        case "euclidean":
            D = torch.cdist(features, features_B, p=2)
        case _:
            raise ValueError("Distance should be 'cosine' or 'euclidean'")
    return D.view((*shape, *D.shape[-2:]))


def affinity_from_features(
    features: torch.Tensor,
    features_B: torch.Tensor = None,
    affinity_focal_gamma: float = 1.0,
    affinity_type: AffinityOptions = "cosine",
):
    """Compute affinity matrix from input features.

    Args:
        features (torch.Tensor): input features, shape (n_samples, n_features)
        features_B (torch.Tensor, optional): optional, if not None, compute affinity between two features
        affinity_focal_gamma (float): affinity matrix parameter, lower t reduce the edge weights
            on weak connections, default 1.0
        affinity_type (str): distance metric, 'cosine' (default) or 'euclidean'.
    Returns:
        (torch.Tensor): affinity matrix, shape (n_samples, n_samples)
    """
    # compute affinity matrix from input features

    # if feature_B is not provided, compute affinity matrix on features x features
    # if feature_B is provided, compute affinity matrix on features x feature_B
    features_B = features if features_B is None else features_B

    # compute distance matrix from input features
    D = distance_from_features(features, features_B, AFFINITY_TO_DISTANCE[affinity_type])

    # lower affinity_focal_gamma reduce the weak edge weights
    match affinity_type:
        case "cosine" | "laplacian":
            A = torch.exp(-D / affinity_focal_gamma)                                        # [... x n x n]
        case "rbf":
            # Outlier-robust scale invariance using quantiles to estimate standard deviation
            c = 2.0
            p = torch.erf(torch.tensor((-c, c), device=features.device) * (2 ** -0.5))
            stds = torch.nanquantile(features, q=(p + 1) / 2, dim=-2)                       # [2 x ... x d]
            stds = (stds[1] - stds[0]) / (2 * c)                                            # [... x d]
            D = 0.5 * (D / torch.norm(stds, dim=-1)[..., None, None]) ** 2
            A = torch.exp(-D / affinity_focal_gamma)
        case _:
            raise ValueError("Affinity should be 'cosine', 'rbf', or 'laplacian'")

    return A



from dataclasses import dataclass
from typing import Literal

import torch
from pytorch3d.ops import sample_farthest_points

from .common import (
    default_device,
)
from .distance_utils import (
    DistanceOptions,
    to_euclidean,
)
from .transformer import (
    TorchTransformerMixin,
)


SampleOptions = Literal["full", "random", "fps", "fps_recursive"]


@dataclass
class SampleConfig:
    method: SampleOptions = "fps"
    num_sample: int = 10000
    fps_dim: int = 12
    n_iter: int = None
    _ncut_obj: TorchTransformerMixin = None


@torch.no_grad()
def subsample_features(
    features: torch.Tensor,
    distance_type: DistanceOptions,
    config: SampleConfig,
):
    features = features.detach()                                                                        # float: [... x n x d]
    with default_device(features.device):
        if config.method == "full" or config.num_sample >= features.shape[0]:
            sampled_indices = torch.arange(features.shape[-2]).expand(features.shape[:-1])              # int: [... x n]
        else:
            # sample
            match config.method:
                case "fps":  # default
                    sampled_indices = fpsample(to_euclidean(features, distance_type), config)

                case "random":  # not recommended
                    mask = torch.all(torch.isfinite(features), dim=-1)                                  # bool: [... x n]
                    weights = mask.to(torch.float) + torch.rand(mask.shape)                             # float: [... x n]
                    sampled_indices = torch.topk(weights, k=config.num_sample, dim=-1).indices          # int: [... x num_sample]

                case "fps_recursive":
                    features = to_euclidean(features, distance_type)                                    # float: [... x n x d]
                    sampled_indices = subsample_features(
                        features=features,
                        distance_type=distance_type,
                        config=SampleConfig(method="fps", num_sample=config.num_sample, fps_dim=config.fps_dim)
                    )                                                                                   # int: [... x num_sample]
                    nc = config._ncut_obj
                    for _ in range(config.n_iter):
                        fps_features, eigenvalues = nc.fit_transform(features, precomputed_sampled_indices=sampled_indices)

                        fps_features = to_euclidean(fps_features[:, :config.fps_dim], "cosine")
                        sampled_indices = torch.sort(fpsample(fps_features, config), dim=-1).values

                case _:
                    raise ValueError("sample_method should be 'farthest' or 'random'")
            sampled_indices = torch.sort(sampled_indices, dim=-1).values
        return sampled_indices


def fpsample(
    features: torch.Tensor,
    config: SampleConfig,
):
    shape = features.shape[:-2]                                                         # ...
    features = features.view((-1, *features.shape[-2:]))                                # [(...) x n x d]
    bsz = features.shape[0]

    mask = torch.all(torch.isfinite(features), dim=-1)                                  # bool: [(...) x n]
    count = torch.sum(mask, dim=-1)                                                     # int: [(...)]
    order = torch.topk(mask.to(torch.int), k=torch.max(count).item(), dim=-1).indices   # int: [(...) x max_count]

    features = torch.nan_to_num(features[torch.arange(bsz)[:, None], order], nan=0.0)   # float: [(...) x max_count x d]
    if features.shape[-1] > config.fps_dim:
        U, S, V = torch.pca_lowrank(features, q=config.fps_dim)                         # float: [(...) x max_count x fps_dim], [(...) x fps_dim], [(...) x fps_dim x fps_dim]
        features = U * S[..., None, :]                                                  # float: [(...) x max_count x fps_dim]

    try:
        sample_indices = sample_farthest_points(
            features, lengths=count, K=config.num_sample
        )[1]                                                                            # int: [(...) x num_sample]
    except RuntimeError:
        original_device = features.device
        alternative_device = "cuda" if original_device == "cpu" else "cpu"
        sample_indices = sample_farthest_points(
            features.to(alternative_device), lengths=count.to(alternative_device), K=config.num_sample
        )[1].to(original_device)                                                        # int: [(...) x num_sample]
    sample_indices = torch.gather(order, 1, sample_indices)                             # int: [(...) x num_sample]

    return sample_indices.view((*shape, *sample_indices.shape[-1:]))                    # int: [... x num_sample]



import einops
import torch

from .nystrom_utils import (
    EigSolverOptions,
    OnlineKernel,
    OnlineNystromSubsampleFit,
    solve_eig,
)
from ..distance_utils import (
    AffinityOptions,
    AFFINITY_TO_DISTANCE,
    affinity_from_features,
)
from ..sampling_utils import (
    SampleConfig,
)


class LaplacianKernel(OnlineKernel):
    def __init__(
        self,
        affinity_focal_gamma: float,
        affinity_type: AffinityOptions,
        adaptive_scaling: bool,
        eig_solver: EigSolverOptions,
    ):
        self.affinity_focal_gamma = affinity_focal_gamma
        self.affinity_type: AffinityOptions = affinity_type
        self.adaptive_scaling: bool = adaptive_scaling
        self.eig_solver: EigSolverOptions = eig_solver

        # Anchor matrices
        self.anchor_features: torch.Tensor = None                                   # [... x n x d]
        self.anchor_mask: torch.Tensor = None
        self.A: torch.Tensor = None                                                 # [... x n x n]
        self.Ainv: torch.Tensor = None                                              # [... x n x n]

        # Updated matrices
        self.a_r: torch.Tensor = None                                               # [... x n]
        self.b_r: torch.Tensor = None                                               # [... x n]

    def fit(self, features: torch.Tensor) -> None:
        self.anchor_features = features                                             # [... x n x d]
        self.anchor_mask = torch.all(torch.isnan(self.anchor_features), dim=-1)     # [... x n]

        self.A = torch.nan_to_num(affinity_from_features(
            self.anchor_features,                                                   # [... x n x d]
            affinity_focal_gamma=self.affinity_focal_gamma,
            affinity_type=self.affinity_type,
        ), nan=0.0)                                                                 # [... x n x n]
        d = features.shape[-1]
        U, L = solve_eig(
            self.A,
            num_eig=d + 1,  # d * (d + 3) // 2 + 1,
            eig_solver=self.eig_solver,
        )                                                                           # [... x n x (d + 1)], [... x (d + 1)]
        self.Ainv = U @ torch.diag_embed(1 / L) @ U.mT                              # [... x n x n]
        self.a_r = torch.where(self.anchor_mask, torch.inf, torch.sum(self.A, dim=-1))  # [... x n]
        self.b_r = torch.zeros_like(self.a_r)                                       # [... x n]

    def _affinity(self, features: torch.Tensor) -> torch.Tensor:
        B = torch.where(self.anchor_mask[..., None], 0.0, affinity_from_features(
            self.anchor_features,                                                   # [... x n x d]
            features,                                                               # [... x m x d]
            affinity_focal_gamma=self.affinity_focal_gamma,
            affinity_type=self.affinity_type,
        ))                                                                          # [... x n x m]
        if self.adaptive_scaling:
            diagonal = (
                einops.rearrange(B, "... n m -> ... m 1 n")                         # [... x m x 1 x n]
                @ self.Ainv                                                         # [... x n x n]
                @ einops.rearrange(B, "... n m -> ... m n 1")                       # [... x m x n x 1]
            ).squeeze(-2, -1)                                                       # [... x m]
            adaptive_scale = diagonal ** -0.5                                       # [... x m]
            B = B * adaptive_scale[..., None, :]
        return B                                                                    # [... x n x m]

    def update(self, features: torch.Tensor) -> torch.Tensor:
        B = self._affinity(features)                                                # [... x n x m]
        b_r = torch.sum(torch.nan_to_num(B, nan=0.0), dim=-1)                       # [... x n]
        b_c = torch.sum(B, dim=-2)                                                  # [... x m]
        self.b_r = self.b_r + b_r                                                   # [... x n]

        row_sum = self.a_r + self.b_r                                               # [... x n]
        col_sum = b_c + (B.mT @ (self.Ainv @ self.b_r[..., None]))[..., 0]          # [... x m]
        scale = (row_sum[..., :, None] * col_sum[..., None, :]) ** -0.5             # [... x n x m]
        return (B * scale).mT                                                       # [... x m x n]

    def transform(self, features: torch.Tensor = None) -> torch.Tensor:
        row_sum = self.a_r + self.b_r                                               # [... x n]
        if features is None:
            B = self.A                                                              # [... x n x n]
            col_sum = row_sum                                                       # [... x n]
        else:
            B = self._affinity(features)
            b_c = torch.sum(B, dim=-2)                                              # [... x m]
            col_sum = b_c + (B.mT @ (self.Ainv @ self.b_r[..., None]))[..., 0]      # [... x m]
        scale = (row_sum[..., :, None] * col_sum[..., None, :]) ** -0.5             # [... x n x m]
        return (B * scale).mT                                                       # [... x m x n]


class NCut(OnlineNystromSubsampleFit):
    """Nystrom Normalized Cut for large scale graph."""

    def __init__(
        self,
        n_components: int = 100,
        affinity_focal_gamma: float = 1.0,
        affinity_type: AffinityOptions = "cosine",
        adaptive_scaling: bool = False,
        sample_config: SampleConfig = SampleConfig(),
        eig_solver: EigSolverOptions = "svd_lowrank",
        chunk_size: int = 8192,
    ):
        """
        Args:
            n_components (int): number of top eigenvectors to return
            affinity_focal_gamma (float): affinity matrix temperature, lower t reduce the not-so-connected edge weights,
                smaller t result in more sharp eigenvectors.
            distance (str): distance metric for affinity matrix, ['cosine', 'euclidean', 'rbf'].
            adaptive_scaling (bool): whether to scale off-diagonal affinity vectors so extended diagonal equals 1
            sample_config (str): subgraph sampling, ['farthest', 'random'].
                farthest point sampling is recommended for better Nystrom-approximation accuracy
            eig_solver (str): eigen decompose solver, ['svd_lowrank', 'lobpcg', 'svd', 'eigh'].
            chunk_size (int): chunk size for large-scale matrix multiplication
        """
        OnlineNystromSubsampleFit.__init__(
            self,
            n_components=n_components,
            kernel=LaplacianKernel(affinity_focal_gamma, affinity_type, adaptive_scaling, eig_solver),
            distance_type=AFFINITY_TO_DISTANCE[affinity_type],
            sample_config=sample_config,
            eig_solver=eig_solver,
            chunk_size=chunk_size,
        )



import torch

from .nystrom_utils import (
    EigSolverOptions,
    OnlineKernel,
    OnlineNystromSubsampleFit,
    solve_eig,
)
from ..distance_utils import (
    DistanceOptions,
    distance_from_features,
)
from ..sampling_utils import (
    SampleConfig,
)


class GramKernel(OnlineKernel):
    def __init__(
        self,
        distance_type: DistanceOptions,
        eig_solver: EigSolverOptions,
    ):
        self.distance_type: DistanceOptions = distance_type
        self.eig_solver: EigSolverOptions = eig_solver

        # Anchor matrices
        self.anchor_features: torch.Tensor = None               # [n x d]
        self.A: torch.Tensor = None                             # [n x n]
        self.Ainv: torch.Tensor = None                          # [n x n]

        # Updated matrices
        self.a_r: torch.Tensor = None                           # [n]
        self.b_r: torch.Tensor = None                           # [n]
        self.matrix_sum: torch.Tensor = torch.zeros(())         # []
        self.n_features: int = None                             # N

    def fit(self, features: torch.Tensor) -> None:
        self.anchor_features = features                         # [n x d]
        self.A = -0.5 * distance_from_features(
            self.anchor_features,                               # [n x d]
            self.anchor_features,
            distance_type=self.distance_type,
        )                                                       # [n x n]
        d = features.shape[-1]
        U, L = solve_eig(
            self.A,
            num_eig=d + 1,  # d * (d + 3) // 2 + 1,
            eig_solver=self.eig_solver,
        )                                                       # [n x (d + 1)], [d + 1]
        self.Ainv = U @ torch.diag(1 / L) @ U.mT                # [n x n]
        self.a_r = torch.sum(self.A, dim=-1)                    # [n]
        self.b_r = torch.zeros_like(self.a_r)                   # [n]
        self.matrix_sum = torch.sum(self.a_r)                   # []
        self.n_features = features.shape[0]                     # n

    def update(self, features: torch.Tensor) -> torch.Tensor:
        B = -0.5 * distance_from_features(
            self.anchor_features,                               # [n x d]
            features,                                           # [m x d]
            distance_type=self.distance_type,
        )                                                       # [n x m]
        b_r = torch.sum(B, dim=-1)                              # [n]
        b_c = torch.sum(B, dim=-2)                              # [m]
        self.b_r = self.b_r + b_r                               # [n]
        self.matrix_sum = (
            torch.sum(self.a_r)
            + 2 * torch.sum(self.b_r)
            + self.Ainv @ self.b_r @ self.b_r
        )                                                       # []
        self.n_features += features.shape[0]                    # N

        row_sum = self.a_r + self.b_r                           # [n]
        col_sum = b_c + B.mT @ self.Ainv @ self.b_r             # [m]
        shift = -(row_sum[:, None] + col_sum) / self.n_features + self.matrix_sum / (self.n_features ** 2)  # [n x m]
        return (B + shift).mT                                   # [m x n]

    def transform(self, features: torch.Tensor = None) -> torch.Tensor:
        row_sum = self.a_r + self.b_r
        if features is None:
            B = self.A                                          # [n x n]
            col_sum = row_sum                                   # [n]
        else:
            B = -0.5 * distance_from_features(
                self.anchor_features,
                features,
                distance_type=self.distance_type,
            )
            b_c = torch.sum(B, dim=-2)                          # [m]
            col_sum = b_c + B.mT @ self.Ainv @ self.b_r         # [m]
        shift = -(row_sum[:, None] + col_sum) / self.n_features + self.matrix_sum / (self.n_features ** 2)  # [n x m]
        return (B + shift).mT                                   # [m x n]


class DistanceRealization(OnlineNystromSubsampleFit):
    """Nystrom Distance Realization for large scale graph."""

    def __init__(
        self,
        n_components: int = 100,
        distance_type: DistanceOptions = "cosine",
        sample_config: SampleConfig = SampleConfig(),
        eig_solver: EigSolverOptions = "svd_lowrank",
        chunk_size: int = 8192,
    ):
        """
        Args:
            n_components (int): number of top eigenvectors to return
            sample_config (str): subgraph sampling, ['farthest', 'random'].
                farthest point sampling is recommended for better Nystrom-approximation accuracy
            distance (str): distance metric for affinity matrix, ['cosine', 'euclidean', 'rbf'].
            eig_solver (str): eigen decompose solver, ['svd_lowrank', 'lobpcg', 'svd', 'eigh'].
            chunk_size (int): chunk size for large-scale matrix multiplication
        """
        OnlineNystromSubsampleFit.__init__(
            self,
            n_components=n_components,
            kernel=GramKernel(distance_type, eig_solver),
            distance_type=distance_type,
            sample_config=sample_config,
            eig_solver=eig_solver,
            chunk_size=chunk_size,
        )

    def fit_transform(
        self,
        features: torch.Tensor,
        precomputed_sampled_indices: torch.Tensor = None,
    ) -> torch.Tensor:
        V = OnlineNystromSubsampleFit.fit_transform(self, features, precomputed_sampled_indices)
        return V * (self.eigenvalues_ ** 0.5)

    def transform(self, features: torch.Tensor = None) -> torch.Tensor:
        V = OnlineNystromSubsampleFit.transform(self, features)
        return V * (self.eigenvalues_ ** 0.5)
